/* Marmote is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Marmote is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Marmote. If not, see <http://www.gnu.org/licenses/>.

Copyright 2015 Alain Jean-Marie, Jean-Michel Fourneau, Jean-Marc Vincent, Issam Rabhi */

#ifndef MARKOVCHAIN_H
#define MARKOVCHAIN_H

#include "marmoteConstants.h"
#include "Distribution/Distribution.h"
#include "simulationResult.h"
#include "transitionStructure/transitionStructure.h"
#include <string>
#include <iostream>

using namespace std;

/**
 * @brief Markov Chain class
 *
 */
class markovChain
{

 protected:

  timeType 		_type; /**< @brief time type: discrete or continuous */
  int 			_stateSpaceSize; /**< @brief size of the state space (should be a pointer on the state space itself) */
  transitionStructure* 	_generator; /**< @brief transition structure of the chain */
  // someType _reward; to be defined later
  // int 			_initState;
  discreteDistribution* _initDistribution; /**< @brief initial distribution of the process */
  bool			_debug; /**< @brief internal debugging indicator */

  bool _isAbstract; /**< @brief true if the object is "abstract", i.e. a pointer to some files */
  int _abstractNbre; /**< @brief number of abstraction parameters */
  string* _abstract; /**< @brief table of abstraction parameters */
  string _format; /**< @brief format/language of the model */
  string _modelName; /**< @brief name of the model */

 public:
  // constructors

  /**
   * @brief Simple constructor for the Markov chain from the size
   * @author Alain Jean-Marie
   * @param sz the size of the state space (may be infinite)
   * @param t the type of chain: CONTINUOUS or DISCRETE
   * @return Markov Chain with no internal structures initialized
   */
  markovChain(int sz, timeType t);


  /**
   * @brief Constructor for the Markov chain using a transition structure.
   * @author Alain Jean-Marie
   * @param tr the transition structure
   * @return Markov Chain with no internal structures initialized
   */

  markovChain(transitionStructure* tr);

  /**
   * @brief Constructor for Markov chains from files in various formats.
   * In the abstract form, the object just stores the name(s) of the files
   * that define the mode. In the non-abstract (concrete) form, the chain is
   * instantiated in the memory with a concrete transition structure.
   * Only the ERS, PSI and Xborne formats are supported at this time for concrete
   * chains.
   * @author Issam Rabhi and Alain Jean-Marie
   * @param format the format or language in which the model is specified
   * @param param[] is the list of parameters
   * @param nbreParam the size of param
   * @param modelName the name of the model, usually the prefix for various files
   * @param isAbstract specifies if the chain is abstract or not
   * @return Markov Chain
   */

  markovChain(string format, string param[], int nbreParam, string modelName,
          bool isAbstract);

  /**
   * @brief Standard destructor. The generator and the initial distrib are destroyed.
   * @author Alain Jean-Marie
   */

  virtual ~markovChain();

public:
  // accessors

  /**
   * @brief Read accessor to get the number of states in the state space of the Markov chain.
   * @return the size of the state space
   */
  int stateSpaceSize() { return _stateSpaceSize; }
  /**
    * @brief Read accessor to get the value of _generator which is a transitionStructure
    *
    * @return the transition structure
    */
  transitionStructure* generator() { return _generator; }
  /**
   * @brief Write accessor to set the value of _initDistribution which is a discreteDistribution
   *
   * @param d the distribution to be set
   */
  void setInitDistribution( discreteDistribution* d ) { if ( _initDistribution != NULL ){ delete(_initDistribution); }; _initDistribution = d; }
  /**
    * @brief Write accessor to set the value of _generator which is a transitionStructure
    *
    * @param tr the transition structure to be set
    */
  void setGenerator( transitionStructure* tr ) { _generator = tr; }

  /**
   * @brief Method(s) for deserializing Distribution from Xborne file (.pi).
   * @author Issam Rabhi
   * @return Distribution
  */

Distribution * read();
/**
 * @brief Utility to set the value of _format
 *
 * @param format
 */
void setFormat(string format){_format=format;}
/**
 * @brief  Utility to set the value of _modelName
 *
 * @param modelName
 */
void setModelName(string modelName){_modelName=modelName;}

/**
 * @brief Utility to set the value of _abstractNbre
 *
 * @param abstractNbre the number of abstract parameters to be set
 */
 void setAbstractNbre(int abstractNbre){_abstractNbre=abstractNbre;}

 /**
  * @brief Utility to set the value of the table containing names related to the model: file names, extensions etc.
  * @author Issam Rabhi and Alain Jean-Marie
  * @param abstract table with the strings to be copied to _abstract
  **/

 void setAbstract(string abstract[]);

 /**
  * @brief Utility to get _abstractNbre
  * @author Issam Rabhi
  * @return _abstractNbre which is the length of the table containing names related to the model.
  **/


 int abstractNbre(){return _abstractNbre;}

 /**
  * @brief Utility to get _modelName.
  * @author Issam Rabhi
  * @return _modelName which is the model name.
**/

 string modelName(){return _modelName;}
 /**
  * @brief Utility to get _format.
  * @author Issam Rabhi
  * @return _format which is the format of the Markov Chain (XbornePres, Xborne, mcl,  ...)
  **/
 string format(){return _format;}

 /**
  * @brief Utility to display the value of the table _abstract[].
  * @author Issam Rabhi
  **/

 void abstract()
 {

     for(int i=0; i<_abstractNbre;i++)
     {
      cout<<i<<":"<<_abstract[i];
     }

 }

 public:
  // solution methods
 /**
  * @brief Simulates the evolution of a Markov Chain using the PSI program.
  * This is a front-end function to both discrete-time and the continuous-time
  * simulators.
  * @author	Alain Jean-Marie and Issam Rabhi
  * @param tMax	time until which the Markov chain is simulated
  * @param Stats	indicates whether occupancy statistics are collected and returned
  * @param Traj	indicates whether a trajectory is returned
  * @param withIncrements	indicates whether time increments should be produced
  * @param Print	indicates whether values should be printed along the way
  * @return a structure with the results of the simulation
  */

  virtual simulationResult* simulateChain( double tMax,
                                           bool Stats, bool Traj,
                                           bool withIncrements, bool Print );

 /**
  * @brief Simulates the evolution of a discrete-time Markov Chain
  * @author	Alain Jean-Marie and Issam Rabhi
  * @param tMax	time until which the Markov chain is simulated
  * @param stats	indicates whether occupancy statistics are collected and returned
  * @param traj	indicates whether a trajectory is returned
  * @param trace	indicates whether the trajectory should be printed along the way (on stdout)
  * @return a structure with the results of the simulation
  */

  virtual simulationResult* simulateChainDT( int tMax,
                                             bool stats, bool traj,
                                             bool trace );
 /**
  * @brief Simulates the evolution of a continuous-time Markov Chain
  * @author	Alain Jean-Marie and Issam Rabhi
  * @param tMax	time until which the Markov chain is simulated
  * @param stats	indicates whether occupancy statistics are collected and returned
  * @param traj	indicates whether a trajectory is returned
  * @param withIncrements	indicates whether time increments should be printed
  * @param trace	indicates whether the trajectory should be printed along the way (on stdout)
  * @return a structure with the results of the simulation
  */

  virtual simulationResult* simulateChainCT( double tMax,
                                             bool stats, bool traj,
                                             bool withIncrements,
                                             bool trace );
 /**
  * @brief Simulates the evolution of a Markov Chain. This is a front-end function
  * to the discrete-time simulator. For continuous time, the PSI simulator
  * does not allow to control the simulation horizon.
  * The effect of "Print" is to be handled in the specific simulation function,
  * not here.
  * @author	Alain Jean-Marie and Issam Rabhi
  * @param tMax	time until which the Markov chain is simulated
  * @param stats	indicates whether occupancy statistics are collected and returned
  * @param Traj	indicates whether a trajectory is returned
  * @param Print	indicates whether values should be printed along the way
  * @return a structure with the results of the simulation
  */

  simulationResult* simulatePSI( int tMax,
                                 bool stats, bool Traj,
                                 bool Print );

  /**
   * @brief Entry point for methods Computing the stationary distribution of chains using
   * the iterative method
   * @author Alain Jean-Marie
   * @param progress indicates whether the progress of the iterative method should be displayed
   * @return an approximation of the stationary distribution
   */

  virtual Distribution* stationaryDistribution( bool progress );
  /**
   * @brief Computing the stationary distribution of a CTMC using uniformization
   * and the iterative method
   * @author Alain Jean-Marie
   * @param progress indicates whether the progress of the iterative
   method should be displayed
   * @return an approximation of the stationary distribution
   */
  virtual Distribution* stationaryDistributionCT( bool progress );
  /**
   * @brief Computing the stationary distribution of a DTMC using uniformization
   * and the iterative method
   * @author Alain Jean-Marie and Issam Rabhi
   * @param progress indicates whether the progress of the iterative method should be displayed
   * @return an approximation of the stationary distribution
   */
  virtual Distribution* stationaryDistributionDT( bool progress );

  /**
   * @brief Entry point for methods computing stationary distributions using the
   * GTH method for solving the linear system
   * @author Issam Rabhi and Alain Jean-Marie
   * @return the stationary distribution, up to errors in the solution of the linear system
   */

  Distribution* stationaryDistributionGthLD();

  /**
   * @brief Entry point for methods computing stationary distributions using the
   * SOR method for solving the linear system
   * @author Alain Jean-Marie
   * @return the stationary distribution, up to errors in the solution of the linear system
   */

  Distribution* stationaryDistributionSOR();

  /**
    * @brief Methods to sample from the stationary distribution using backwards coupling.
    * The result is returned in a simulationResult object, but the interpretation is different:
    * it does not represent trajectories.
    * Here, "state" entries are the measured states, and "time" entries are the measured coupling time.
    * @author Alain Jean-Marie
    * @param nbSample number of samples to collect
    * @return an object storing the measured samples and coupling times
    */
  simulationResult* stationaryDistributionSample( int nbSamples );

  /**
    * @brief Entry point for methods computing the distribution of the hitting time (first entry times)
    * from some state to some set of states.
    * @author Alain Jean-Marie
    * @param iState index of the initial state
    * @param hitSetIndicator a boolean array where states in the hitting set are marked with true
    * @return the array of average hitting times from every state not in the hitting set
    */
  Distribution* hittingTimeDistribution( int iState, bool *hitSetIndicator );

  /**
   * @brief Obtain samples of hitting times through Monte Carlo simulation.
   * A trajectory is simulated until it hits the target set, or its length attains a maximum,
   * whichever comes first. Samples with the maximum are returned although they do not represent
   * a proper hitting time. It is the responsibility of the calling party to ignore these values.
   * @param iState the initial state from which trajectories start
   * @param hittingSetIndicator boolean array indicating with true which states are in the target
   * @param nbSamples number of samples to collect
   * @param tMax maximum length of trajectories
   * @return an array with the samples
   */
  int* simulateHittingTime(int iState, bool *hittingSet,
                                            int nbSamples, int tMax );
  /**
    * @brief Entry point for methods computing average hitting times (first entry times) from every state to some set of states.
    * @author Alain Jean-Marie
    * @param hitSetIndicator a boolean array where states in the hitting set are marked with true
    * @return the array of average hitting times from every state not in the hitting set
    */
  double* averageHittingTime( bool *hitSetIndicator );

  /**
    * @brief Computing the average hitting times in a discrete-time Markov chain.
    * Uses a direct Gauss-Seidel matrix inversion.
    * @author Alain Jean-Marie
    * @param hitSetIndicator a boolean array where states in the hitting set are marked with true
    * @return the array of average hitting times from every state not in the hitting set
    */
  double* averageHittingTimeDT( bool *hitSetIndicator );

  /**
    * @brief Computing the average hitting times in a discrete-time Markov chain.
    * Uses an iterative approximate computation.
    * @author Alain Jean-Marie
    * @param hitSetIndicator a boolean array where states in the hitting set are marked with true
    * @return the array of average hitting times from every state not in the hitting set
    */
  double* averageHittingTimeDT_iterative( bool *hitSetIndicator );

 public:

  /**
   * @brief copy utility
   *
   * @return markovChain
   */
  virtual markovChain* copy();


  /**
   * @brief Uniformize Markov Chain, by uniformizing the generator.
   * If the chain is already discrete time, a copy is returned.
   * @author Alain Jean-Marie
   * @return a new Markov chain, uniformized version of the original
   */

  virtual markovChain* uniformize();

  /**
   * @brief Function to find out the size and the type of a Markov chain described
   * in the MARCA format. Both quantities are directly set in the method,
   * which returns nothing.
   * Adapted from the method HBF::read_marka of Psi/Unix/v1.0.
   * @author Alain Jean-Marie
   **/

  void setSizeType(const string path);
  // virtual void makeMarkovChain();

 public:

  /**
   * @brief Method(s) for writing Markov chains in files with various formats.
   * Only the ERS format is supported at this time.
   * @author Alain Jean-Marie
   * @param out the file descriptor in which to write the chain
   * @param withReward specifies if rewards are to be written; not used yet
   */
  virtual void write( FILE *out, bool withReward );
  virtual void write( string format, string modelName);

};

#endif // MARKOVCHAIN_H
